/**CFile***********************************************************************

  FileName    [cuddBddIte.c]

  PackageName [cudd]

  Synopsis    [BDD ITE function and satellites.]

  Description [External procedures included in this module:
		<ul>
                <li> Cudd_bddIte()
       	        <li> Cudd_bddIteConstant() taken out
		<li> Cudd_bddIntersect()
		<li> Cudd_bddAnd()
		<li> Cudd_bddOr()
		<li> Cudd_bddNand()
		<li> Cudd_bddNor()
		<li> Cudd_bddXor()
		<li> Cudd_bddXnor()
		<li> Cudd_bddLeq()
		</ul>
       Internal procedures included in this module:
		<ul>
		<li> cuddBddIteRecur()
		<li> cuddBddIntersectRecur()
		<li> cuddBddAndRecur()
		<li> cuddBddXorRecur()
		</ul>
       Static procedures included in this module:
		<ul>
       	        <li> bddVarToConst()
       	        <li> bddVarToCanonical()
       	        <li> bddVarToCanonicalSimple()
		</ul>

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright   [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include "util.h"
#include "cuddInt.h"
#include "message.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define NONE_RECIEVED 0
#define THEN_RECIEVED 1
#define ELSE_RECIEVED 2
#define BOTH_RECIEVED 3

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddBddIte.c,v 1.9 1997/03/08 18:25:30 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static DdNode*
Cudd_bdd2(DdManager *dd, int op, DdNode* f, DdNode* g);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Existentially abstracts all the variables in cube from f.]

  Description [Existentially abstracts all the variables in cube from f.
  Returns the abstracted BDD if successful; NULL otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_bddUnivAbstract Cudd_addExistAbstract]

******************************************************************************/
DdNode *
Cudd_bddExistAbstract(manager, f, cube)
DdManager	*manager;
DdNode		*f;
DdNode		*cube;
{
  return(Cudd_bdd2(manager,CUDD_BDD_EXISTS,f,cube));
} /* end of Cudd_bddExistAbstract */


/**Function********************************************************************

  Synopsis    [Universally abstracts all the variables in cube from f.]

  Description [Universally abstracts all the variables in cube from f.
  Returns the abstracted BDD if successful; NULL otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_bddExistAbstract Cudd_addUnivAbstract]

******************************************************************************/
DdNode *
Cudd_bddUnivAbstract(manager, f, cube)
DdManager	*manager;
DdNode		*f;
DdNode          *cube;
{
  return(Cudd_bdd2(manager,CUDD_BDD_FORALL,f,cube));
} /* end of Cudd_bddUnivAbstract */

/**Function********************************************************************

  Synopsis    [Implements ITE(f,g,h).]

  Description [Implements ITE(f,g,h). Returns a pointer to the
  resulting BDD if successful; NULL if the intermediate result blows
  up.]

  SideEffects [None]

  SeeAlso     [Cudd_addIte Cudd_bddIteConstant Cudd_bddIntersect]

******************************************************************************/
DdNode*
Cudd_bddIte(dd, f, g, h)
DdManager *dd;
DdNode* f;
DdNode* g;
DdNode* h;
{
  DdNodePar res_par;
  DdNode *one,*zero;
  one=DD_ONE(dd);
  zero=Cudd_Not(one);

  if(f==one)
    return g;
  if(f==zero)
    return h;
  if(g==h)
    return g;

  messageSendIte(dd,cuddNodeUserToPar(dd,f),cuddNodeUserToPar(dd,g),cuddNodeUserToPar(dd,h));
  res_par=DdMessageRecieveMaster(dd);
  return cuddUserNodeInter(dd,res_par);  
} /* end of Cudd_bddIte */


/**Function********************************************************************

  Synopsis [Takes the AND of two BDDs and simultaneously abstracts the
  variables in cube.]

  Description [Takes the AND of two BDDs and simultaneously abstracts
  the variables in cube. The variables are existentially abstracted.
  Returns a pointer to the result is successful; NULL otherwise.
  Cudd_bddAndAbstract implements the semiring matrix multiplication
  algorithm for the boolean semiring.]

  SideEffects [None]

  SeeAlso     [Cudd_addMatrixMultiply Cudd_addTriangle Cudd_bddAnd]

******************************************************************************/
DdNode *
Cudd_bddAndAbstract(dd, f, g, cube)
DdManager *dd;
DdNode    *f;
DdNode    *g;
DdNode    *cube;
{
  DdNodePar f_par,g_par,res_par;

  f_par=cuddNodeUserToPar(dd,f);
  g_par=cuddNodeUserToPar(dd,g);

  if(cuddIsConstantPar(f_par) && cuddIsConstantPar(g_par))
    res_par=cuddBddAndBasePar(dd,f_par,g_par);
  else
    {
      messageSendAndAbstract(dd,f_par,g_par,cuddNodeUserToPar(dd,cube));
      res_par=DdMessageRecieveMaster(dd);
    }
  return cuddUserNodeInter(dd,res_par); 
} /* end of Cudd_bddAndAbstract */


/**Function********************************************************************

  Synopsis    [Computes the conjunction of two BDDs f and g.]

  Description [Computes the conjunction of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAndAbstract Cudd_bddIntersect
  Cudd_bddOr Cudd_bddNand Cudd_bddNor Cudd_bddXor Cudd_bddXnor]

******************************************************************************/
DdNode *
Cudd_bddAnd(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_AND,f,g));
} /* end of Cudd_bddAnd */


/**Function********************************************************************

  Synopsis    [Computes the disjunction of two BDDs f and g.]

  Description [Computes the disjunction of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAnd Cudd_bddNand Cudd_bddNor
  Cudd_bddXor Cudd_bddXnor]

******************************************************************************/
DdNode *
Cudd_bddOr(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_OR,f,g));
} /* end of Cudd_bddOr */


/**Function********************************************************************

  Synopsis    [Computes the NAND of two BDDs f and g.]

  Description [Computes the NAND of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAnd Cudd_bddOr Cudd_bddNor
  Cudd_bddXor Cudd_bddXnor]

******************************************************************************/
DdNode *
Cudd_bddNand(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_NAND,f,g));
} /* end of Cudd_bddNand */


/**Function********************************************************************

  Synopsis    [Computes the NOR of two BDDs f and g.]

  Description [Computes the NOR of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAnd Cudd_bddOr Cudd_bddNand
  Cudd_bddXor Cudd_bddXnor]

******************************************************************************/
DdNode *
Cudd_bddNor(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_NOR,f,g));
} /* end of Cudd_bddNor */


/**Function********************************************************************

  Synopsis    [Computes the exclusive OR of two BDDs f and g.]

  Description [Computes the exclusive OR of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAnd Cudd_bddOr
  Cudd_bddNand Cudd_bddNor Cudd_bddXnor]

******************************************************************************/
DdNode *
Cudd_bddXor(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_XOR,f,g));
} /* end of Cudd_bddXor */


/**Function********************************************************************

  Synopsis    [Computes the exclusive NOR of two BDDs f and g.]

  Description [Computes the exclusive NOR of two BDDs f and g. Returns a
  pointer to the resulting BDD if successful; NULL if the intermediate
  result blows up.]

  SideEffects [None]

  SeeAlso     [Cudd_bddIte Cudd_addApply Cudd_bddAnd Cudd_bddOr
  Cudd_bddNand Cudd_bddNor Cudd_bddXor]

******************************************************************************/
DdNode *
Cudd_bddXnor(dd, f, g)
DdManager *dd;
DdNode *f;
DdNode *g;
{
  return(Cudd_bdd2(dd,CUDD_BDD_XNOR,f,g));
} /* end of Cudd_bddXnor */


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/* generalized two argument Bdd operation */
static DdNode*
Cudd_bdd2(dd, op, f, g)
DdManager *dd;
int op;
DdNode* f;
DdNode* g;
{
    DdNode* res;
    DdNodePar f_par,g_par,res_par;
    f_par=cuddNodeUserToPar(dd,f);
    g_par=cuddNodeUserToPar(dd,g);

    switch(op)
      {
      case CUDD_BDD_OR: case CUDD_BDD_NOR:
	f_par=Cudd_NotPar(f_par);
	g_par=Cudd_NotPar(g_par);
	break;
      case CUDD_BDD_XNOR:
	g_par=Cudd_NotPar(g_par);
	break;
      case CUDD_BDD_FORALL:
	f_par=Cudd_NotPar(f_par);
      }
    switch(op)
      {
      case CUDD_BDD_OR:  case CUDD_BDD_NOR:
      case CUDD_BDD_AND: case CUDD_BDD_NAND:
	res_par=cuddBddAndBasePar(dd,f_par,g_par);
	if(cuddEqualPar(res_par,CUDD_ERROR))
	  {
	    messageSendAnd(dd,f_par,g_par);
	    res_par=DdMessageRecieveMaster(dd);
	  }
	break;
      case CUDD_BDD_XOR: case CUDD_BDD_XNOR:
	res_par=cuddBddXorBasePar(dd,f_par,g_par);
	if(cuddEqualPar(res_par,CUDD_ERROR))
	  {
	    messageSendXor(dd,f_par,g_par);
	    res_par=DdMessageRecieveMaster(dd);
	  }
	break;
      case CUDD_BDD_FORALL: case CUDD_BDD_EXISTS:
	if(cuddIsConstantPar(f_par) || cuddIsConstantPar(g_par))
	  res_par=f_par;
	else
	  {
	    messageSendExists(dd,f_par,g_par);
	    res_par=DdMessageRecieveMaster(dd);
	  }
	break;	
      default:
	assert(0 && (int)"illigal operation");
      }

    switch(op)
      {
      case CUDD_BDD_OR: 
      case CUDD_BDD_NAND:
      case CUDD_BDD_FORALL:
	res_par=Cudd_NotPar(res_par);
	/* note  Cudd_NotPar(CUDD_ERROR)==CUDD_ERROR */
      }

    /* note cuddUserNodeInter is safe */
    res = cuddUserNodeInter(dd,res_par);

    return(res);

} /* end of Cudd_bdd2 */

