#include "cuddInt.h"
#include "message.h"

void ddCacheResizePar(DdManager *table)
{
    DdCache *cache,*oldcache,*entry,*old;
    int i;
    int posn;
    unsigned int slots,oldslots;

    oldcache = table->cache;
    oldslots = table->cacheSlots;
    slots = table->cacheSlots = Cudd_Prime(oldslots * DD_GROW_FACTOR);

#ifdef DD_VERBOSE
    (void) fprintf(stderr,"Resizing the cache(s) from %d to %d entries\n",
		oldslots, slots);
    (void) fprintf(stderr,"\thits = %d\tlookups = %d\thit ratio = %5.3f\n",
		table->cacheHits, table->cacheLookUps,
		(double) table->cacheHits / (double) table->cacheLookUps);
#endif

    table->cache = cache = ALLOC(DdCache,slots);
    table->memused += (slots - oldslots) * sizeof(DdCache);

    /* Clear new cache. */
    for (i = 0; (unsigned) i < slots; i++) {
	cache[i].f = CUDD_ERROR;
	/* note f=CUDD_ERROR now represents open spots */
#ifdef DD_CACHE_PROFILE
	cache[i].count = 0;
#endif
    }

    /* Copy from old cache to new one. */
    for (i = 0; (unsigned) i < oldslots; i++) {
	old = &oldcache[i];
	if (!cuddEqualPar(old->f,CUDD_ERROR)) {
	    posn = ddCHash(old->op,cuddValuePar(old->f),cuddValuePar(old->g),cuddValuePar(old->h),slots);
	    entry = &cache[posn];
	    entry->op = old->op;
	    entry->f = old->f;
	    entry->g = old->g;
	    entry->h = old->h;
	    entry->data = old->data;	
	}
    }

    FREE(oldcache);

    /* Reinitialize measurements so as to avoid division by 0 and
    ** immediate resizing.
    */
    table->cacheLookUps = slots * DD_MIN_HIT / 100 + 1;
    table->cacheHits = 0;
}


void cuddCacheWait(DdManager *manager,DdCacheHandle handle,DdCallTreeNode caller)
{
  DdCacheWaitingList en;
  en=ALLOC(struct DdCacheWaitingList,1);
  /* NOTE diffrent memory management should be considered */
  en->caller=caller;
  en->next=handle.entry->WaitingList;
  handle.entry->WaitingList=en;
} 

void cuddCacheReactivate(DdManager *manager,DdCacheWaitingList l,DdNodePar data)
{
  DdCacheWaitingList tmp;
  
  while(l!=NULL)
    {
      DdCallTreeUpdateNode(manager,l->caller,data);
      tmp=l;
      l=l->next;
      free(tmp);
    }
}

void cuddCacheUpdatePar(DdManager *table, DdCache *entry, DdNodePar data)
{
    if(entry!=NULL)
      {
	table->cacheinserts++;
	
	entry->data = data;
	
	cuddCacheReactivate(table,entry->WaitingList,data);
      }
#ifdef DD_CACHE_PROFILE
    entry->count++;
#endif
}

/* Returns to position en the cache where the entry is, or would have been
   if the space is occupied by work in progress. if the entry is available,
   it is updated to show work in progress. */
DdCacheHandle cuddCacheCheck2Par(DdManager *table, int op, DdNodePar f, DdNodePar g)
{
    DdCacheHandle res;
    int posn;
    register DdCache *entry,*cache;

    cache = table->cache;

    /* use the symmerty in most binary operations to increase
       cache efficiency 
       Also use the fact that f.index<=g.index (to get order when
       f.address = g.address) */
    if(cuddAdrOrder(f.address,g.address) && op != CUDD_BDD_EXISTS)
      {
	DdNodePar tmp = f;
	f = g;
	g = tmp;
      }

    table->cacheLookUps++;
    posn = ddCHash2(op,cuddValuePar(f),cuddValuePar(g),table->cacheSlots);
    entry = &cache[posn];

    if(cuddEqualPar(entry->f,f) && cuddEqualPar(entry->g,g) && entry->op==op) 
      { 
	res.entry=entry;
	table->totCachehits++;
	table->cacheHits++;
	if(cuddEqualPar(entry->data,CUDD_WAIT))
	  res.status= CUDD_CACHE_IN_PROGRESS;
	else
	  res.status= CUDD_CACHE_HIT;
      }
    else
      {
	res.status=CUDD_CACHE_MISS;	
	if(cuddEqualPar(entry->f,CUDD_ERROR) ||
	   !cuddEqualPar(entry->data,CUDD_WAIT)) 
	  { /* entry can be grabbed */
	    if(!cuddEqualPar(entry->f,CUDD_ERROR))
	      table->cachecollisions++;
	    entry->op = op;
	    entry->f = f;
	    entry->g = g;
	    entry->h = g;
	    entry->data = CUDD_WAIT;
	    entry->WaitingList=NULL;
	    res.entry=entry;
	  }
	else
	  res.entry=NULL;

	/* decide whether to resize */
	/* Rev3: Resize only at flush desicition made when flush up is done */
      }
    return res;
}

DdCacheHandle cuddCacheCheck3Par(DdManager *table, int op, DdNodePar f, DdNodePar g,DdNodePar h)
{
    DdCacheHandle res;
    int posn;
    register DdCache *entry,*cache;

    cache = table->cache;

    /* bring the operation to canocial form to increase
       cache efficiency */
    if(op == CUDD_BDD_ITE)
      {
	/* ITE(!f,g,h) == ITE(f,h,g) */
	if(Cudd_IsComplementPar(f))
	  {
	    DdNodePar tmp = g;
	    g = h;
	    h = tmp;
	    f=Cudd_NotPar(f);
	  }

	/*  ignoring ITE(f,!g,h) == !ITE(f,g,!h) 
	    because the optimization requires keeping track of 
	    when results needs complement, especially in connection
	    with waiting cache hits */
      }
    else if(op == CUDD_BDD_ANDABSTRACT)
      {
	/* f and g are symetric */
	if(cuddAdrOrder(f.address,g.address))
	  {
	    DdNodePar tmp = f;
	    f = g;
	    g = tmp;
	  }
	/* ignoring the possible misses from f.address = g.address */ 	
      }
    else
      assert(0); /* ITE and ANDABSTRACT atr the only 3 parameter functions */


    table->cacheLookUps++;
    posn = ddCHash(op,cuddValuePar(f),cuddValuePar(g),cuddValuePar(h),table->cacheSlots);
    entry = &cache[posn];

    if(cuddEqualPar(entry->f,f) && cuddEqualPar(entry->g,g) && cuddEqualPar(entry->h,h) && entry->op==op) 
      { 
	res.entry=entry;
	table->totCachehits++;
	table->cacheHits++;
	if(cuddEqualPar(entry->data,CUDD_WAIT))
	  res.status= CUDD_CACHE_IN_PROGRESS;
	else
	  res.status= CUDD_CACHE_HIT;
      }
    else
      {
	res.status=CUDD_CACHE_MISS;	
	if(cuddEqualPar(entry->f,CUDD_ERROR) ||
	   !cuddEqualPar(entry->data,CUDD_WAIT)) 
	  { /* entry can be grabbed */
	    if(!cuddEqualPar(entry->f,CUDD_ERROR))
	      table->cachecollisions++;
	    entry->op = op;
	    entry->f = f;
	    entry->g = g;
	    entry->h = h;
	    entry->data = CUDD_WAIT;
	    entry->WaitingList=NULL;
	    res.entry=entry;
	  }
	else
	  res.entry=NULL;

	/* decide whether to resize */
	/* Rev3: Resize only at flush desicition made when flush up is done */
      }
    return res;
}

DdCacheHandle cuddCacheCheckItePar(DdManager *manager, int op, DdNodePar f, DdNodePar g,DdNodePar h)
{
    switch(op)
    {
    case CUDD_BDD_OR_1_3:
    case CUDD_BDD_OR_NOT1_2:
      /* ignoring that or can be done as and,
	 to avoid negating on the way up */
    case CUDD_BDD_ITE:
      return cuddCacheCheck3Par(manager,CUDD_BDD_ITE,f,g,h);
      break;
    case CUDD_BDD_AND_1_2:
      return cuddCacheCheck2Par(manager,CUDD_BDD_AND,f,g);
      break;
    case CUDD_BDD_AND_NOT1_3:
      return cuddCacheCheck2Par(manager,CUDD_BDD_AND,Cudd_NotPar(f),h);
      break;
    }
}
