/**CFile***********************************************************************

  FileName    [cuddInit.c]

  PackageName [cudd]

  Synopsis    [Functions to initialize and shut down the DD manager.]

  Description [External procedures included in this module:
		<ul>
		<li> Cudd_Init()
		<li> Cudd_Quit()
		</ul>
	       Internal procedures included in this module:
		<ul>
		<li> cuddZddInitUniv()
		<li> cuddZddFreeUniv()
		</ul>
	      ]

  SeeAlso     []

  Author      [Fabio Somenzi]

  Copyright [ This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

  Rev 2   Changes include:
              Cudd_Init works with DdNodePars
              Declarations of functions from cudd.h (which had been macros)
                   Cudd_NotPar
                   Cudd_ComplementPar

******************************************************************************/

#include "message.h"
#include    "util.h"
#define CUDD_MAIN
#include    "cuddInt.h"
#undef CUDD_MAIN

/*---------------------------------------------------------------------------*/
/* Declarations from cudd.h                                                  */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

#ifndef lint
static char rcsid[] DD_UNUSED = "$Id: cuddInit.c,v 1.17 1997/02/16 21:42:25 fabio Exp $";
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int cuddInitConstants(DdManager *unique);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Creates a new DD manager.]

  Description [Creates a new DD manager, initializes the table, the
  basic constants and the projection functions. If maxCacheSize is 0,
  Cudd_Init decides a suitable value based on the available
  memory. Returns a pointer to the manager if successful; NULL
  otherwise.]

  SideEffects [None]

  SeeAlso     [Cudd_Quit]

******************************************************************************/
DdManager *
Cudd_Init(numVars,numVarsZ,numSlots,cacheSize,maxCacheSize)
unsigned int numVars;	/* Initial number of BDD variables (i.e., subtables) */
unsigned int numVarsZ;	/* Initial number of ZDD variables (i.e., subtables) */
                        /* kept for compability reasons */
unsigned int numSlots;	/* Initial size of the unique tables */
unsigned int cacheSize;	/* Initial size of the cache */
unsigned int maxCacheSize; /* Size of the cache beyond which no resizinig occurs */
{
    DdManager *unique;
    int i,result;

    assert(numVarsZ==0);

    unique = cuddInitTablePar(ddMax(DD_DEFAULT_RESIZE, numVars));
    if (unique == NULL) return(NULL);
    /*
    if (maxCacheSize == 0) {
	long datasize = getSoftDataLimit();
	maxCacheSize = (datasize / sizeof(DdCache)) / 4;
    }
    */

    result = cuddInitCachePar(unique,cacheSize,maxCacheSize);
    if (result == 0) return(NULL);

    result = cuddInitUserNodeTable(unique);
    if (result == 0) return(NULL);

    result = cuddInitConstants(unique);
    if (result == 0) return(NULL);
            
    /* init communication */
    DdInitMaster(unique);
    InitProcesses(unique,cacheSize,maxCacheSize);

    /* Create the projection functions. */
    /* Rev 3: only reserve space, by default no vars to start */
#ifdef __osf__
#pragma pointer_size save
#pragma pointer_size short
#endif
    unique->vars = ALLOC(DdNode * ,unique->maxSize);
#ifdef __osf__
#pragma pointer_size restore
#endif
    unique->memused += sizeof(DdNode *) * unique->maxSize;
    
    return(unique);

} /* end of Cudd_Init */


/**Function********************************************************************

  Synopsis    [Deletes resources associated with a DD manager.]

  Description [Deletes resources associated with a DD manager and
  resets the global statistical counters. (Otherwise, another manaqger
  subsequently created would inherit the stats of this one.)]

  SideEffects [None]

  SeeAlso     [Cudd_Init]

******************************************************************************/
void
Cudd_Quit(unique)
DdManager *unique;
{
  /*    cuddFreeTablePar(unique); */
    int i;
    DdNode *node;
    
    /*
    for(i=1;i<=unique->UserNodes.keys;i++)
      {
	node=&unique->UserNodes.PageTable[0][i];
	printf("node:%8x index %2d ref: %2d NodePar: %2d,%5d,%2d\n",(unsigned)node,node->index,node->ref,node->NodePar.index,node->NodePar.address.page,node->NodePar.address.number);
      }
      */

    KillProcesses(unique);
      

#ifdef DD_STATS
    /* REV2 nuked
    cudd_Nodesfreed = 0;
    cudd_Nodesdropped = 0;
    */
#ifdef DD_UNIQUE_PROFILE
    cudd_UniqueLookUps = 0;
    cudd_UniqueLinks = 0;
#endif
#endif
    pvm_exit();
} /* end of Cudd_Quit */


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

static int cuddInitConstants(DdManager *unique)
{
  unique->one.index   = CUDD_MAXINDEX;
  /* NOTE the contents of the one node is never read, so it need not
     actually be stored anywhere, and there is no memory for constants
     nodes */
  unique->one.address.page   = 0;
  unique->one.address.number = 1; /* numbers are 1 indexed */

  /* note UserNodeInter does not work when index==CUDD_MAXINDEX */
  unique->oneUser=cuddAllocNodeUser(unique);
  if(unique->oneUser==NULL)
    return(0);
  unique->oneUser->index = CUDD_MAXINDEX;
  unique->oneUser->ref = DD_MAXREF;
  unique->oneUser->NodePar = unique->one;
  unique->oneUser->next=NULL_USER;

  /* place oneUser in hash table */
  unique->UserNodes.nodelist[ddHashUser(unique->one,unique->UserNodes.slots)]=
    unique->oneUser;
  unique->UserNodes.keys++;
  return(1);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/









